from cython cimport size_t
from libcpp.string cimport string
from libcpp.unordered_map cimport unordered_map
from libcpp.vector cimport vector

cdef extern from "wordpiece.hh":
    cdef cppclass Piece:
        string piece
        bint is_initial
        Piece()
        Piece(const string& piece, bint initial)

    cdef cppclass PieceStorage:
        vector[Piece] _id_to_piece

        PieceStorage()
        void add_piece(const string &piece, const bint is_initial)
        size_t size()
        const Piece &id_to_piece(const int id) except +
        size_t piece_to_id(const Piece &piece) except +
        int try_piece_to_id(const Piece &piece)




from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.utility cimport pair

cdef extern from "merges.hh":
    cdef cppclass Merges:
        Merges(const vector[pair[string, string]]& merges)
        vector[string] apply_merges(vector[string] pieces) const
        vector[pair[string, string]] merges() const


from cutlery import ByteBPEProcessor
import json
from pathlib import Path
import pytest


@pytest.fixture(scope="module")
def test_dir(request):
    return Path(request.fspath).parent


@pytest.fixture
def toy_processor(test_dir):
    return ByteBPEProcessor.load_from_files(
        vocab=test_dir / "robbert-vocab-1000.json",
        merges=test_dir / "robbert-merges-1000.txt",
    )


EXAMPLE_TEXT = " Wij bezoeken alle provinciën."
EXAMPLE_PIECES = [
    "ĠWij",
    "Ġbez",
    "oeken",
    "Ġalle",
    "Ġpro",
    "v",
    "in",
    "c",
    "iÃ«n",
    ".",
]

EXAMPLE_PIECE_IDS = [280, 3377, 7095, 92, 1951, 253, 194, 311, 3698, 4]


def test_empty_processor():
    bbpe = ByteBPEProcessor({}, [])
    # An empty processor never merges, only recodes bytes.
    assert bbpe.encode_as_pieces("they'll visit Köln") == [
        "t",
        "h",
        "e",
        "y",
        "'",
        "l",
        "l",
        "Ġ",
        "v",
        "i",
        "s",
        "i",
        "t",
        "Ġ",
        "K",
        "Ã",
        "¶",
        "l",
        "n",
    ]

def test_can_decode(toy_processor):
    assert toy_processor.decode_from_ids(EXAMPLE_PIECE_IDS) == EXAMPLE_TEXT


def test_can_encode(toy_processor):
    assert toy_processor.encode(EXAMPLE_TEXT) == (EXAMPLE_PIECE_IDS, EXAMPLE_PIECES)


def test_can_encode_as_ids(toy_processor):
    assert toy_processor.encode_as_ids(EXAMPLE_TEXT) == EXAMPLE_PIECE_IDS


def test_can_encode_as_pieces(toy_processor):
    assert toy_processor.encode_as_pieces(EXAMPLE_TEXT) == EXAMPLE_PIECES


def test_rejects_incorrect_merges(test_dir):
    with pytest.raises(ValueError, match=r"Merge must consist of 2 items"):
        ByteBPEProcessor.load_from_files(
            vocab=test_dir / "robbert-vocab-1000.json",
            merges=test_dir / "incorrect-merges.txt",
        )



from typing import List, Tuple, Iterable
from cython.operator cimport dereference as deref
from libcpp cimport pair

cdef struct PieceMatch:
    bint found
    int prefix_len
    int piece_id

cdef class WordPieceProcessor:
    cdef PieceStorage _pieces

    def __init__(self, pieces: List[str]):
        self._pieces = PieceStorage()

        for idx, piece in enumerate(pieces):
            is_initial = not piece.startswith("##")
            byte_array = piece[2:].encode('utf8') if not is_initial else piece.encode('utf8')
            self._pieces.add_piece(byte_array, is_initial)

    def encode(self, token: str) -> Tuple[List[int], List[str]]:
        """
        Encode a token using word pieces. The piece identifiers are
        returned. If no piece could be found for a suffix, the special
        piece identifier `-1` is used.

            token (str): The token to encode as pieces.
            RETURNS (Tuple[List[int], List[str]]): Piece identifiers and token pieces.
        """
        token_ids = []
        token_pieces = []
        prefix = ""
        cdef PieceMatch match
        is_initial = True
        while token:
            match = self._find_longest_prefix(token, is_initial)
            if not match.found:
                token_ids.append(-1)
                token_pieces.append(None)
                return token_ids, token_pieces
            token_ids.append(match.piece_id)
            token_pieces.append(f"{prefix}{token[:match.prefix_len]}")
            token = token[match.prefix_len:]
            is_initial = False
            prefix = "##"
        return token_ids, token_pieces

    def decode(self, pieces: Iterable[int]) -> str:
        """
        Decode token piece identifiers into string. Raises a `RuntimeError` 
        if any of the identifiers are invalid.

            ids (Iterable[int]): Piece IDs.
            RETURNS (str): Decoded string.
        """
        token_pieces = [self.id_to_piece(id)[0] for id in pieces]
        return "".join(token_pieces)

    def get_initial(self, piece: str) -> int:
        """
        Returns the ID for the given initial piece. Raises a `RuntimeError` if 
        the string isn't an initial piece.

            piece (str): Initial piece string.
            RETURNS (int): Piece ID.
        """
        return self._pieces.piece_to_id(Piece(piece.encode('utf8'), True))

    def id_to_piece(self, piece: int) -> Tuple[str, bool]:
        """
        Returns the piece string (without any prefix) for a given piece identifier 
        and a boolean identifying if it is an initial piece. Raises a `RuntimeError` if 
        any of the identifiers are invalid.

            piece (int): Piece ID.
            RETURNS (Tuple[str, bool]): Piece string and a boolean identifying if
                it is an initial piece.
        """
        cdef const Piece* ptr_piece = &self._pieces.id_to_piece(piece)
        return (ptr_piece[0].piece.decode('utf8'), ptr_piece[0].is_initial) 

    def is_valid_piece_id(self, piece: int) -> bool:
        """
        Returns True if the piece identifier is valid, False otherwise.

            piece (int): Piece ID.
            RETURNS (bool): Is valid.
        """
        return 0 <= piece <= self._pieces.size() - 1

    @staticmethod
    def from_file(filename: str) -> WordPieceProcessor:
        with open(filename, encoding="utf8") as f:
            pieces = [line.strip() for line in f]
        return WordPieceProcessor(pieces)

    def to_list(self):
        pieces = [None] * self._pieces.size()
        cdef Piece piece
        for i in range(self._pieces.size()):
            piece = self._pieces._id_to_piece[i]
            if piece.is_initial:
                pieces[i] = piece.piece.decode('utf8')
            else:
                pieces[i] = f"##{piece.piece.decode('utf8')}"

        return pieces

    cdef PieceMatch _find_longest_prefix(self, str token, bint is_initial):
        cdef PieceMatch match
        match.found = False

        cdef Piece piece
        for prefix_len in reversed(range(len(token) + 1)): 
            piece = Piece(token[:prefix_len].encode("utf8"), is_initial)
            idx = self._pieces.try_piece_to_id(piece)
            if idx != -1:
                match.found = True
                match.prefix_len = prefix_len
                match.piece_id = idx
                return match

        return match


from pathlib import Path
import pytest

from cutlery import SentencePieceProcessor


@pytest.fixture(scope="module")
def test_dir(request):
    return Path(request.fspath).parent


@pytest.fixture
def toy_model(test_dir):
    return SentencePieceProcessor.from_file(str(test_dir / "toy.model"))


def test_load_proto(test_dir):
    with open(str(test_dir / "toy.model"), "rb") as f:
        data = f.read()
    spp = SentencePieceProcessor.from_protobuf(data)
    _check_ids(spp)
    serialized_data = spp.to_protobuf()
    assert serialized_data == data


def test_load_unknown_file():
    with pytest.raises(OSError, match=r"No such file"):
        SentencePieceProcessor.from_file("bogus.model")


def test_handles_nul_character(toy_model):
    ids, pieces = toy_model.encode("Test\0 nul")
    assert ids == [239, 382, 0, 7, 24, 231]
    assert pieces == ["▁T", "est", "\0", "▁", "n", "ul"]


def test_decode_from_ids(toy_model):
    decoded = toy_model.decode_from_ids(
        [8, 465, 10, 947, 41, 10, 170, 168, 110, 28, 20, 143, 4]
    )
    assert decoded == "I saw a girl with a telescope."


def test_decode_from_pieces(toy_model):
    decoded = toy_model.decode_from_pieces(
        [
            "▁I",
            "▁saw",
            "▁a",
            "▁girl",
            "▁with",
            "▁a",
            "▁t",
            "el",
            "es",
            "c",
            "o",
            "pe",
            ".",
        ]
    )
    assert decoded == "I saw a girl with a telescope."


def test_decode_with_pieces_rejects_inccorect_type(toy_model):
    with pytest.raises(TypeError):
        toy_model.decode_from_pieces("test")
    with pytest.raises(TypeError):
        toy_model.decode_from_pieces([1, 2, 3])


def test_encode(toy_model):
    ids, pieces = toy_model.encode("I saw a girl with a telescope.")
    assert ids == [8, 465, 10, 947, 41, 10, 170, 168, 110, 28, 20, 143, 4]
    assert pieces == [
        "▁I",
        "▁saw",
        "▁a",
        "▁girl",
        "▁with",
        "▁a",
        "▁t",
        "el",
        "es",
        "c",
        "o",
        "pe",
        ".",
    ]


def test_encode_as_ids(toy_model):
    _check_ids(toy_model)


def test_encode_as_pieces(toy_model):
    pieces = toy_model.encode_as_pieces("I saw a girl with a telescope.")
    assert pieces == [
        "▁I",
        "▁saw",
        "▁a",
        "▁girl",
        "▁with",
        "▁a",
        "▁t",
        "el",
        "es",
        "c",
        "o",
        "pe",
        ".",
    ]


def test_uninitialized_model():
    spp = SentencePieceProcessor()
    with pytest.raises(RuntimeError):
        spp.encode("I saw a girl with a telescope.")
    with pytest.raises(RuntimeError):
        spp.encode_as_ids("I saw a girl with a telescope.")
    with pytest.raises(RuntimeError):
        spp.encode_as_pieces("I saw a girl with a telescope.")
    with pytest.raises(RuntimeError):
        spp.decode_from_pieces(["▁I"])
    with pytest.raises(RuntimeError):
        spp.decode_from_ids([8])
    with pytest.raises(RuntimeError):
        spp.bos_id()
    with pytest.raises(RuntimeError):
        spp.eos_id()
    with pytest.raises(RuntimeError):
        spp.unk_id()
    with pytest.raises(RuntimeError):
        spp.pad_id()


def _check_ids(spp):
    assert len(spp) == 1000
    assert spp.bos_id() == 1
    assert spp.eos_id() == 2
    assert spp.unk_id() == 0
    assert spp.pad_id() == -1  # Disabled in this model.
    ids = spp.encode_as_ids("I saw a girl with a telescope.")
    assert ids == [8, 465, 10, 947, 41, 10, 170, 168, 110, 28, 20, 143, 4]


def test_id_to_piece_and_piece_to_id(toy_model):
    assert toy_model.piece_to_id("<s>") == toy_model.bos_id()
    assert toy_model.piece_to_id("</s>") == toy_model.eos_id()
    assert toy_model.piece_to_id("<unk>") == toy_model.unk_id()
    assert toy_model.piece_to_id("qotsa") == toy_model.unk_id()

    assert toy_model.id_to_piece(toy_model.bos_id()) == "<s>"
    assert toy_model.id_to_piece(toy_model.eos_id()) == "</s>"
    assert toy_model.id_to_piece(toy_model.unk_id()) == "<unk>"
    with pytest.raises(ValueError):
        toy_model.id_to_piece(-1)
    with pytest.raises(ValueError):
        toy_model.id_to_piece(len(toy_model))



from cython.operator cimport dereference as deref
from typing import List, Tuple

cdef class SentencePieceProcessor:
    def __cinit__(self):
        self.spp.reset(new CSentencePieceProcessor())

    def __init__(self):
        pass

    def __len__(self):
        _check_status(deref(self.spp).status())
        return deref(self.spp).GetPieceSize()

    @staticmethod
    def from_file(str filename):
        """
        Constructs a SentencePieceProcessor by loading a serialized model from
        disk.

            filename (str): Path to model.
        """
        cdef SentencePieceProcessor processor = SentencePieceProcessor.__new__(SentencePieceProcessor)
        _check_status(deref(processor.spp).Load(filename.encode("utf-8")))
        return processor

    @staticmethod
    def from_protobuf(bytes protobuf):
        """
        Constructs a SentencePieceProcessor by loading a serialized model from
        a protocol buffer.

            protobuf (bytes): Byte array representing the model's serialized
                protocol buffer.
        """
        cdef SentencePieceProcessor processor = SentencePieceProcessor.__new__(SentencePieceProcessor)
        cdef string_view protobuf_view = string_view(protobuf, len(protobuf))
        _check_status(deref(processor.spp).LoadFromSerializedProto(protobuf_view))
        return processor

    def to_protobuf(self) -> bytes:
        """
        Serializes the SentencePieceProcessor to a protocol buffer.

            RETURNS (bytes): Byte array representing the model's serialized
                protocol buffer.
        """
        cdef string serialized = deref(self.spp).serialized_model_proto()
        return bytes(serialized)

    def decode_from_ids(self, list ids) -> str:
        """
        Decodes piece indentifiers to a string. Raises a `ValueError` if the
        list elements are not integers.

            ids (List[int]): Piece IDs.
            RETURNS (str): Decoded string.
        """
        cdef vector[int] c_ids
        for piece_id in ids:
            if not isinstance(piece_id, int):
                raise TypeError("Pieces must be of type `int` when decoding from ids")
            c_ids.push_back(piece_id)
        cdef string output
        _check_status(deref(self.spp).Decode(c_ids, &output))
        return output.decode("utf-8")

    def decode_from_pieces(self, list pieces) -> str:
        """
        Decodes piece tokens to a string. Raises a `ValueError` if the
        list elements are not strings.

            ids (List[str]): Piece tokens.
            RETURNS (str): Decoded string.
        """
        cdef vector[string] c_strings
        for piece in pieces:
            if not isinstance(piece, str):
                raise TypeError("Pieces must be of type `str` when decoding from pieces")
            c_strings.push_back(piece.encode("utf-8"))
        cdef string output
        _check_status(deref(self.spp).Decode(c_strings, &output))
        return output.decode("utf-8")

    def encode(self, str sentence) -> Tuple[List[int], List[str]]:
        """
        Encodes a string to piece indentifiers and piece tokens.

            sentence (str): Text to encode.
            RETURNS (Tuple[List[int], List[str]]): Piece IDs and piece tokens
                corresponding to the input text.
        """
        cdef SentencePieceText text = self._encode(sentence)

        cdef int idx
        cdef SentencePieceText_SentencePiece piece
        pieces = []
        ids = []
        for idx in range(text.pieces_size()):
            piece = text.pieces(idx)
            ids.append(piece.id())
            pieces.append(piece.piece().decode("utf-8"))

        return ids, pieces

    def encode_as_ids(self, str sentence) -> List[int]:
        """
        Encodes a string to piece indentifiers.

            sentence (str): Text to encode.
            RETURNS (List[int]): Piece IDs.
        """
        cdef SentencePieceText text = self._encode(sentence)

        cdef int idx
        ids = []
        for idx in range(text.pieces_size()):
            ids.append(text.pieces(idx).id())

        return ids

    def encode_as_pieces(self, str sentence) -> List[str]:
        """
        Encodes a string to piece tokens.

            sentence (str): Text to encode.
            RETURNS (List[str]): Piece tokens.
        """
        cdef SentencePieceText text = self._encode(sentence)

        cdef int idx
        cdef SentencePieceText_SentencePiece piece
        pieces = []
        for idx in range(text.pieces_size()):
            piece = text.pieces(idx)
            pieces.append(text.pieces(idx).piece().decode("utf-8"))

        return pieces

    def piece_to_id(self, str piece) -> int:
        """
        Returns the piece identifier for a given piece token. Returns the 
        `unk` piece identifier if the piece token is OOV.

            piece (str): Piece token.
            RETURNS (int): Piece ID.
        """
        piece_bytes = piece.encode("utf8")
        cdef string_view piece_view = string_view(piece_bytes, len(piece_bytes))
        _check_status(deref(self.spp).status())
        return deref(self.spp).PieceToId(piece_view)

    def id_to_piece(self, int piece_id) -> str:
        """
        Returns the piece token for a given piece identifier. Raises a `ValueError` 
        if the piece identifier is out-of-bounds.

            piece_id (int): Piece ID.
            RETURNS (str): Piece token.
        """
        _check_status(deref(self.spp).status())
        cdef int vocab_size = deref(self.spp).GetPieceSize()
        if not 0 <= piece_id < vocab_size:
            raise ValueError(f"piece ID must be in range [0,{vocab_size}), got {piece_id}")
        return deref(self.spp).IdToPiece(piece_id).decode("utf8")

    def bos_id(self) -> int:
        """Returns the piece identifier for the `bos` meta token."""
        _check_status(deref(self.spp).status())
        return deref(self.spp).bos_id()

    def eos_id(self) -> int:
        """Returns the piece identifier for the `eos` meta token."""
        _check_status(deref(self.spp).status())
        return deref(self.spp).eos_id()

    def pad_id(self) -> int:
        """Returns the piece identifier for the `pad` meta token."""
        _check_status(deref(self.spp).status())
        return deref(self.spp).pad_id()

    def unk_id(self) -> int:
        """Returns the piece identifier for the `unk` meta token."""
        _check_status(deref(self.spp).status())
        return deref(self.spp).unk_id()

    cdef SentencePieceText _encode(self, str sentence) except * :
        sentence_bytes = sentence.encode("utf-8")
        cdef SentencePieceText text;
        _check_status(deref(self.spp).Encode(sentence.encode("utf-8"), &text))
        return text

cdef _check_status(Status status):
    cdef code = <int> status.code()
    if code == <int> kOk:
        return
    elif code == <int> kNotFound:
        raise OSError(status.error_message().decode("utf-8"))
    elif code == <int> kInternal:
        raise RuntimeError(status.error_message().decode("utf-8"))
    else:
        raise ValueError(status.error_message().decode("utf-8"))



# cython: infer_types

from typing import Dict, Iterable, List, Optional, Tuple
from cython.operator cimport dereference as deref
from functools import lru_cache
from io import BytesIO
import json
from libcpp.memory cimport make_shared, shared_ptr
from pathlib import Path
import regex


# We need the same splitting pattern and byte -> string remapping as GPT-2,
# so these are taken directly from:
#
# https://github.com/openai/gpt-2/blob/master/src/encoder.py
#
# See the LICENSE file for licensing information.


SPLIT_PATTERN = r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+"""


@lru_cache()
def bytes_to_unicode() -> Dict[int, bytes]:
    """
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = (
        list(range(ord("!"), ord("~") + 1))
        + list(range(ord("¡"), ord("¬") + 1))
        + list(range(ord("®"), ord("ÿ") + 1))
    )
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n).encode("utf-8") for n in cs]
    return dict(zip(bs, cs))

cdef class ByteBPEProcessor:
    cdef shared_ptr[Merges] _merges
    cdef dict _byte_decoder
    cdef dict _byte_encoder
    cdef object _split_pattern
    cdef dict _piece_to_id
    cdef dict _id_to_piece

    def __init__(self, vocab: Dict[str, int], merges: List[Tuple[str, str]]):
        self._byte_encoder = bytes_to_unicode()
        self._byte_decoder = {v.decode("utf-8"): k for k, v in self._byte_encoder.items()}
        self._split_pattern = regex.compile(SPLIT_PATTERN)
        self._piece_to_id = vocab
        self._id_to_piece = {v: k for k, v in vocab.items()}
        cdef vector[pair[string, string]] c_merges
        for p1, p2 in merges:
            c_merges.push_back(pair[string, string](p1.encode('utf-8'), p2.encode('utf-8')))
        self._merges = make_shared[Merges](c_merges)

    @staticmethod
    def load_from_files(*, vocab: Path, merges: Path) -> ByteBPEProcessor:
        """Construct a processor from the given vocabulary and merges files."""
        with open(vocab, encoding="utf-8") as f:
            vocab = json.load(f)
        with open(merges, encoding="utf-8") as f:
            version = f.readline()
            if not version.startswith("#version: 0.2"):
                raise ValueError(f"Only version 0.2 of the merges format is supported, was: {version.strip()}")
            merges = []
            for line in f:
                merge = line.strip().split(" ")
                if len(merge) != 2:
                    raise ValueError(f"Merge must consist of 2 items, was {len(merge)}: {line.strip()}")
                merges.append(merge)
        return ByteBPEProcessor(vocab, merges)

    def decode_from_ids(self, ids: Iterable[int]) -> str:
        """
        Decode piece identifiers into string.

            ids (Iterable[int]): piece identifiers.
            RETURNS (str): decoded string.
        """
        decoded = BytesIO()
        for piece_id in ids:
            piece = self._id_to_piece.get(piece_id, None)
            if piece is None:
                raise ValueError(f"Unknown piece identifier: {piece_id}")
            decoded.write(bytes([self._byte_decoder[cp] for cp in list(piece)]))

        return decoded.getvalue().decode("utf-8")

    def encode(self, text: str) -> Tuple[List[int], List[str]]:
        """
        Split a text into pieces.

            text: The text to split.
            RETURNS: A pair of pieces and piece identifiers.
        """
        pieces = self.encode_as_pieces(text)
        return self._pieces_to_ids(pieces), pieces

    def encode_as_ids(self, text: str) -> List[int]:
        """
        Split a text into pieces, returning piece identifiers.

            text: The text to split.
            RETURNS: Piece identifiers.
        """
        pieces = self.encode_as_pieces(text)
        return self._pieces_to_ids(pieces)

    def encode_as_pieces(self, text: str) -> List[str]:
        """
        Split a text into pieces, returning piece pieces.

            text: The text to split.
            RETURNS: Piece pieces.
        """
        pieces = []
        for token in regex.findall(self._split_pattern, text):
            token_bytes = [self._byte_encoder[b] for b in token.encode("utf-8")]
            merged = deref(self._merges).apply_merges(token_bytes)
            pieces.extend(merge.decode("utf-8") for merge in merged)
        return pieces

    @property
    def merges(self) -> List[Tuple[str, str]]:
        """Get all merges."""
        merges_bytes = deref(self._merges).merges()
        return [(m1.decode('utf-8'), m2.decode('utf-8')) for m1, m2 in merges_bytes]

    def piece_id(self, piece: str) -> Optional[int]:
        """Get the identifier for a piece."""
        return self._piece_to_id.get(piece)

    @property
    def vocab(self) -> Dict[str, int]:
        """Get a copy of the vocabulary."""
        return dict(self._piece_to_id)

    def _pieces_to_ids(self, pieces):
        piece_ids = []
        for piece in pieces:
            piece_id = self._piece_to_id.get(piece)
            if piece_id is None:
                raise ValueError(f"Piece is not in vocabulary: {piece}")
            piece_ids.append(piece_id)

        return piece_ids



from libc.stdint cimport uint32_t
from libcpp.memory cimport shared_ptr
from libcpp.string cimport string
from libcpp.vector cimport vector

cdef extern from "builtin_pb/sentencepiece.pb.h" namespace "sentencepiece":
    cdef cppclass SentencePieceText_SentencePiece:
        uint32_t id() const
        const string & piece() const

    cdef cppclass SentencePieceText:
        const SentencePieceText_SentencePiece& pieces(int index) const;
        int pieces_size() const


cdef extern from "sentencepiece_processor.h" namespace "sentencepiece":
    cdef cppclass CSentencePieceProcessor "sentencepiece::SentencePieceProcessor":
        SentencePieceProcessor()
        string EncodeAsSerializedProto(const string& filename)
        Status Decode(const vector[int]& ids, string *detokenized)
        Status Decode(const vector[string]& ids, string *detokenized)
        Status Encode(const string& input, SentencePieceText * spt)
        int GetPieceSize() const
        Status Load(const string& filename)
        Status LoadFromSerializedProto(string_view serialized);
        Status status() const
        int bos_id() const
        int eos_id() const
        int pad_id() const
        int unk_id() const
        string serialized_model_proto() const
        int PieceToId(string_view piece) const
        const string &IdToPiece(int id) const


cdef extern from "sentencepiece_processor.h" namespace "absl":
    cdef cppclass string_view:
        string_view()
        string_view(const char * data, size_t len)


cdef extern from "sentencepiece_processor.h" namespace "sentencepiece::util":
    # Cython 3.0 has support for C++ enum classes. Until then, work around it.
    cdef cppclass StatusCode:
        pass

    cdef cppclass Status:
        StatusCode code() const
        const char * error_message() const

cdef extern from "sentencepiece_processor.h" namespace "sentencepiece::util::StatusCode":
    cdef StatusCode kOk
    cdef StatusCode kInternal
    cdef StatusCode kNotFound

cdef class SentencePieceProcessor:
    cdef shared_ptr[CSentencePieceProcessor] spp

    cdef SentencePieceText _encode(self, str sentence) except *



from cutlery import WordPieceProcessor
from pathlib import Path
import pytest


EXAMPLE_TOKENS = [
    "voor",
    "voorman",
    "coördinatie",
    "voorkomen",
]

EXAMPLE_ENCODINGS = [
    ([0], ["voor"]),
    ([0, -1], ["voor", None]),
    ([2, 1], ["coördina", "##tie"]),
    ([0, 3, 4], ["voor", "##kom", "##en"]),
]

EXAMPLE_DECODINGS = [
    "voor",
    "",  # Will raise an error.
    "coördinatie",
    "voorkomen",
]

TOKEN_PIECES = ["voor", "##tie", "coördina", "##kom", "##en"]


@pytest.fixture(scope="module")
def test_dir(request):
    return Path(request.fspath).parent


@pytest.fixture
def toy_processor_from_file(test_dir):
    return WordPieceProcessor.from_file(str(test_dir / "toy-word-pieces.txt"))


@pytest.fixture
def toy_processor():
    return WordPieceProcessor(TOKEN_PIECES)


def test_word_piece_processor_small_encode(toy_processor):
    for token, output in zip(EXAMPLE_TOKENS, EXAMPLE_ENCODINGS):
        y = toy_processor.encode(token)
        assert y == output


def test_word_piece_processor_small_decode(toy_processor):
    for encoding, result in zip(EXAMPLE_ENCODINGS, EXAMPLE_DECODINGS):
        if -1 in encoding[0]:
            with pytest.raises(RuntimeError):
                toy_processor.decode(encoding[0])
        else:
            assert toy_processor.decode(encoding[0]) == result


def test_to_list(toy_processor):
    assert toy_processor.to_list() == TOKEN_PIECES


def test_from_file(toy_processor_from_file):
    assert toy_processor_from_file.to_list() == TOKEN_PIECES


def test_get_initial(toy_processor):
    assert toy_processor.get_initial("voor") == 0
    with pytest.raises(RuntimeError):
        toy_processor.get_initial("##tie")


def test_piece_id_valid(toy_processor):
    assert toy_processor.is_valid_piece_id(-1) == False
    assert toy_processor.is_valid_piece_id(0) == True
    assert toy_processor.is_valid_piece_id(len(TOKEN_PIECES)) == False
    assert toy_processor.is_valid_piece_id(len(TOKEN_PIECES) - 1) == True



#!/usr/bin/env python
import sys
import distutils.util
from distutils.command.build_ext import build_ext
from setuptools import Extension, setup, find_packages
from pathlib import Path
from Cython.Build import cythonize
from Cython.Compiler import Options


# Preserve `__doc__` on functions and classes
# http://docs.cython.org/en/latest/src/userguide/source_files_and_compilation.html#compiler-options
Options.docstrings = True


def prefix_path(prefix, files):
    return list(map(lambda f: f"{prefix}/{f}", files))


ABSL_SRC = prefix_path(
    "sentencepiece/third_party/absl", ["flags/flag.cc", "strings/string_view.cc"]
)

PROTOBUF_LIGHT_SRC = prefix_path(
    "sentencepiece/third_party/protobuf-lite",
    [
        "arena.cc",
        "arenastring.cc",
        "bytestream.cc",
        "coded_stream.cc",
        "common.cc",
        "extension_set.cc",
        "generated_enum_util.cc",
        "generated_message_table_driven_lite.cc",
        "generated_message_util.cc",
        "implicit_weak_message.cc",
        "int128.cc",
        "io_win32.cc",
        "message_lite.cc",
        "parse_context.cc",
        "repeated_field.cc",
        "status.cc",
        "statusor.cc",
        "stringpiece.cc",
        "stringprintf.cc",
        "structurally_valid.cc",
        "strutil.cc",
        "time.cc",
        "wire_format_lite.cc",
        "zero_copy_stream.cc",
        "zero_copy_stream_impl.cc",
        "zero_copy_stream_impl_lite.cc",
    ],
)

SENTENCEPIECE_PROTOBUF_SRC = prefix_path(
    "sentencepiece/src/builtin_pb", ["sentencepiece.pb.cc", "sentencepiece_model.pb.cc"]
)

SENTENCEPIECE_SRC = prefix_path(
    "sentencepiece/src",
    [
        "bpe_model.cc",
        "char_model.cc",
        "error.cc",
        "filesystem.cc",
        "model_factory.cc",
        "model_interface.cc",
        "normalizer.cc",
        "sentencepiece_processor.cc",
        "unigram_model.cc",
        "util.cc",
        "word_model.cc",
    ],
)

PACKAGES = find_packages()
MOD_NAMES = [
    "cutlery.sp",
]
COMPILE_OPTIONS = {
    "msvc": [
        "/Ox",
        "/EHsc",
        "/D_USE_INTERNAL_STRING_VIEW",
        "/DHAVE_PTHREAD",
        "/wd4018",
        "/wd4514",
    ],
    "other": [
        "--std=c++14",
        "-Wno-sign-compare" "-Wno-strict-prototypes",
        "-Wno-unused-function",
        "-D_USE_INTERNAL_STRING_VIEW",
        "-pthread",
        "-DHAVE_PTHREAD=1",
    ],
}
COMPILER_DIRECTIVES = {
    "language_level": -3,
    "embedsignature": True,
    "annotation_typing": False,
}
LINK_OPTIONS = {"msvc": [], "other": []}


def is_new_osx():
    """Check whether we're on OSX >= 10.10"""
    name = distutils.util.get_platform()
    if sys.platform != "darwin":
        return False
    elif name.startswith("macosx-10"):
        minor_version = int(name.split("-")[1].split(".")[1])
        if minor_version >= 7:
            return True
        else:
            return False
    else:
        return False


if is_new_osx():
    # On Mac, use libc++ because Apple deprecated use of libstdc
    COMPILE_OPTIONS["other"].append("-stdlib=libc++")
    LINK_OPTIONS["other"].append("-lc++")
    # g++ (used by unix compiler on mac) links to libstdc++ as a default lib.
    # See: https://stackoverflow.com/questions/1653047/avoid-linking-to-libstdc
    LINK_OPTIONS["other"].append("-nodefaultlibs")


# By subclassing build_extensions we have the actual compiler that will be used
# which is really known only after finalize_options
# http://stackoverflow.com/questions/724664/python-distutils-how-to-get-a-compiler-that-is-going-to-be-used
class build_ext_options:
    def build_options(self):
        if hasattr(self.compiler, "initialize"):
            self.compiler.initialize()
        self.compiler.platform = sys.platform[:6]
        for e in self.extensions:
            e.extra_compile_args = COMPILE_OPTIONS.get(
                self.compiler.compiler_type, COMPILE_OPTIONS["other"]
            )
            e.extra_link_args = LINK_OPTIONS.get(
                self.compiler.compiler_type, LINK_OPTIONS["other"]
            )


class build_ext_subclass(build_ext, build_ext_options):
    def build_extensions(self):
        build_ext_options.build_options(self)
        build_ext.build_extensions(self)


def clean(path):
    for path in path.glob("**/*"):
        if path.is_file() and path.suffix in (".so", ".cpp"):
            print(f"Deleting {path.name}")
            path.unlink()


def setup_package():
    root = Path(__file__).parent

    if len(sys.argv) > 1 and sys.argv[1] == "clean":
        return clean(root / "cutlery")

    ext_modules = [
        Extension(
            "cutlery._spp",
            ["cutlery/_spp.pyx"]
            + ABSL_SRC
            + PROTOBUF_LIGHT_SRC
            + SENTENCEPIECE_SRC
            + SENTENCEPIECE_PROTOBUF_SRC,
            include_dirs=[
                "cutlery",
                "sentencepiece",
                "sentencepiece/src",
                "sentencepiece/src/builtin_pb",
                "sentencepiece/third_party/protobuf-lite",
            ],
            language="c++",
        ),
        Extension(
            "cutlery._wordpiece",
            ["cutlery/_wordpiece.pyx", "cutlery/wordpiece.cc"],
            language="c++",
        ),
        Extension(
            "cutlery._bbpe",
            ["cutlery/_bbpe.pyx", "cutlery/merges.cc"],
            include_dirs=["cutlery"],
            language="c++",
        ),
    ]
    print("Cythonizing sources")
    ext_modules = cythonize(
        ext_modules, compiler_directives=COMPILER_DIRECTIVES, language_level=2
    )
    setup(
        name="cutlery",
        packages=PACKAGES,
        ext_modules=ext_modules,
        cmdclass={"build_ext": build_ext_subclass},
        package_data={"": ["*.pyx", "*.pxd", "*.pxi", "*.cu", "*.hh"]},
    )


if __name__ == "__main__":
    setup_package()
